from typing import Any, Union, List, Optional

import io


class HttpResponse(object):
    def __init__(self, content: Union[str, io.StringIO], content_type: str = Optional[None]) -> None: ...


class HttpResponsePermanentRedirect(HttpResponse): ...

class HttpResponseRedirect(HttpResponse): ...

class Http404(Exception): ...



# Stubs for django.shortcuts (Python 3.6)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from django.db import models

from typing import Any, Optional, Union, TypeVar, Generic, Type
from typing_extensions import Protocol

T = TypeVar('T', bound=models.Model)

class SupportsAbsoluteURL(Protocol):
    def get_absolute_url(self) -> str:
       ...

from django.http import HttpResponsePermanentRedirect, HttpResponseRedirect

# def render_to_response(template_name: Any, context: Optional[Any] = ..., content_type: Optional[Any] = ..., status: Optional[Any] = ..., using: Optional[Any] = ...): ...
# def render(request: Any, template_name: Any, context: Optional[Any] = ..., content_type: Optional[Any] = ..., status: Optional[Any] = ..., using: Optional[Any] = ...): ...
def redirect(to: Union[str, SupportsAbsoluteURL], *, permanent: bool = False) -> Union[HttpResponseRedirect, HttpResponsePermanentRedirect]: ... # *args: Any, permanent: bool = ..., **kwargs: Any): ...
def get_object_or_404(klass: Type[T], *args: Any, **kwargs: Any) -> T: ...
# def get_list_or_404(klass: Any, *args: Any, **kwargs: Any): ...
# def resolve_url(to: Any, *args: Any, **kwargs: Any): ...



from typing import Any, Optional

class Model:
    id: Optional[int]
    pk: Optional[int]

    def save(self, *args: Any, **kwargs: Any) -> None: ...

CharField: Any
BooleanField: Any
ForeignKey: Any
TextField: Any
SlugField: Any
IntegerField: Any
ManyToManyField: Any
DateTimeField: Any
EmailField: Any

CASCADE: Any



from django import forms
from django.conf import settings
from django.contrib import messages
from django.http import HttpRequest, HttpResponse
from django.middleware.csrf import get_token
from django.template.defaultfilters import escape

from typing import Any, Dict, Tuple, Union, Sequence, Mapping, TypeVar, Callable, Type, overload, Optional, cast, List, NamedTuple

from mypy_extensions import TypedDict, Arg, KwArg

import abc
import datetime
import simplejson


type_registry: Dict[str, Tuple] = {}


Serializable = Tuple[
    Union[
        None,
        str,
        bool,
        Dict[str, Union[str, int, float, bool, None]],
        forms.BaseForm,
        Sequence[
            Tuple[
                Union[
                    Sequence[
                        Tuple[
                            Union[
                                str,
                                bool,
                                int,
                                None,
                            ],
                            ...
                        ],
                    ],
                    str,
                    bool,
                    int,
                    Dict[str, Union[str, int, float, bool, None]],
                ],
                ...
            ]
        ],

        Tuple[
            Union[
                str,
                int,
                float,
                bool,

                Sequence[
                    Tuple[
                        Union[
                            str,
                            int,
                            float,
                            bool,
                            'TypeHint',
                            None,
                        ],
                        ...
                    ]
                ],

                Mapping[str, Union[
                    str,
                    int,
                    float,
                    bool,
                    Sequence[str],
                    None
                ]],
                'TypeHint',
            ],
            ...
        ]
    ],
    ...
]


K = TypeVar('K')
P = TypeVar('P', bound=Serializable)

View = Callable[[HttpRequest, K], Union[P, HttpResponse]]
NoArgsView = Callable[[HttpRequest], Union[P, HttpResponse]]


def to_camel_case(snake_str: str) -> str:
    components = snake_str.split('_')
    return ''.join(x.title() for x in components)


    reveal_type(form_view)


class Message(NamedTuple):
    level: int
    level_tag: str
    message: str


class Request(NamedTuple):
    path: str


class Context(NamedTuple):
    template_name: str
    csrf_token: str
    request: Request
    messages: List[Message]


def encode_complex_types(obj: Any) -> Serializable:
    """
    Handles dates, forms, and other types.
    From: https://stackoverflow.com/a/22238613
    """
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    elif isinstance(obj, forms.BaseForm):
        return serialize_form(obj)

    raise TypeError("Type %s not serializable" % type(obj))


class JSXResponse:
    def __init__(self, *, context: Context, props: P) -> None:

        self.data = {
            'context': context._asdict(),
            'props': props if isinstance(props, dict) else props._asdict(),  # type: ignore
        }

    def as_json(self) -> Any:
        return simplejson.dumps(self.data, default=encode_complex_types)  #, use_decimal=False)


def render_jsx(request: HttpRequest, template_name: str, props: Union[P, HttpResponse]) -> HttpResponse:
    if isinstance(props, HttpResponse):
        return props


    current_messages = messages.get_messages(request)

    response = JSXResponse(
        context=Context(
            request=Request(
                path=request.path,
            ),
            template_name=template_name,
            csrf_token=get_token(request),
            messages=[
                Message(
                    level=m.level,
                    level_tag=m.level_tag,
                    message=m.message,
                ) for m in current_messages
            ],
        ),
        props=props,
    )

    if 'debug' in request.GET:
        return HttpResponse('<html><body><h1>Debug response</h1><pre>' + escape(response.as_json()) + '</pre></body></html>', content_type='text/html')

    return HttpResponse(response.as_json(), content_type='application/ssr+json')


@overload
def ssr(*,
        props: Type[P],
        params: None = None) -> Callable[[NoArgsView[P]], Callable[[Arg(HttpRequest, 'request'), KwArg(Any)], HttpResponse]]: ...


@overload
def ssr(*,
        props: Type[P],
        params: Type[K]) -> Callable[[View[K, P]], Callable[[Arg(HttpRequest, 'request'), KwArg(Any)], HttpResponse]]: ...


def ssr(*,
        props: Type[P],
        params: Optional[Type[K]] = None) -> Union[
                                                 Callable[[NoArgsView[P]], Callable[[Arg(HttpRequest, 'request'), KwArg(Any)], HttpResponse]],
                                                 Callable[[View[K, P]], Callable[[Arg(HttpRequest, 'request'), KwArg(Any)], HttpResponse]],
                                             ]:
    type_registry[props.__name__] = props  # type: ignore

    def no_args_wrap_with_jsx(original: NoArgsView[P]) -> Callable[[Arg(HttpRequest, 'request'), KwArg(Any)], HttpResponse]:
        def wrapper(request: HttpRequest, **kwargs: Any) -> HttpResponse:
            props = original(request)
            template_name = to_camel_case(original.__name__)

            return render_jsx(request, template_name, props)

        return wrapper

    def wrap_with_jsx(original: View[K, P]) -> Callable[[Arg(HttpRequest, 'request'), KwArg(Any)], HttpResponse]:
        def wrapper(request: HttpRequest, **kwargs: Any) -> HttpResponse:
            props = original(request, cast(Any, params)(**kwargs))
            template_name = to_camel_case(original.__name__)

            return render_jsx(request, template_name, props)
        return wrapper

    if params is None:
        return no_args_wrap_with_jsx
    else:
        return wrap_with_jsx


class TypeHint(abc.ABC):
    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass


def create_schema(Type: Any, definitions: Dict, ref: bool = True) -> Any:
    if (getattr(Type, '__origin__', None) == Union or
        str(Type.__class__) == 'typing.Union'):  # TODO: find a better way to do this.
        return {
            'anyOf': [
                create_schema(field, definitions) for field in Type.__args__
            ],
        }
    elif str(Type.__class__) == 'typing.Any':  # TODO: find a better way to do this.
        return {
        }
    elif getattr(Type, '_name', None) == 'Dict':
        return {
            'type': 'object',
            'additionalProperties': create_schema(Type.__args__[1], definitions),
        }
    elif getattr(Type, '_name', None) == 'List':
        return {
            'type': 'array',
            'items': create_schema(Type.__args__[0], definitions),
        }
    elif issubclass(Type, List):
        return {
            'type': 'array',
            'items': create_schema(Type.__args__[0], definitions),
        }
    elif issubclass(Type, Dict):
        return {
            'type': 'object',
            'additionalProperties': create_schema(Type.__args__[1], definitions),
        }
    elif issubclass(Type, bool):
        return {
            'type': 'bool',
        }
    elif issubclass(Type, int):
        return {
            'type': 'number',
        }
    elif issubclass(Type, str):
        return {
            'type': 'string',
        }
    elif Type is type(None):
        return {
            'type': 'null',
        }
    elif hasattr(Type, '_asdict'):
        definition_name = f'{Type.__module__}.{Type.__qualname__}'

        if ref is False or definition_name not in definitions:
            required = []
            properties = {}

            for field_name, SubType in Type.__annotations__.items():
                field_schema = create_schema(SubType, definitions)

                if field_schema is not None:

                    required.append(field_name)
                    properties[field_name] = field_schema

            definition = {
                'title': Type.__name__,
                'type': 'object',
                'additionalProperties': False,
                'properties': properties,
                'required': required,
            }

            if ref is False:
                return definition
            definitions[definition_name] = definition

        return {
            '$ref': f'#/definitions/{definition_name}',
        }
    elif issubclass(Type, forms.BaseForm):
        definition_name = f'{Type.__module__}.{Type.__qualname__}'
        required = []
        properties = {}
        field_schema = create_schema(FieldType, definitions)
        # We manually build errors using type augmentation.
        # error_schema = create_schema(FormErrors, definitions)

        for field_name, SubType in Type.base_fields.items():
            required.append(field_name)
            properties[field_name] = field_schema

        definitions['Form'] = {
            'title': 'Form',
            'type': 'object',
            'additionalProperties': False,
        }
        definitions[definition_name] = {
            'title': Type.__name__,
            'allOf': [
                {
                    '$ref': '#/definitions/Form',
                },
                {
                    'type': 'object',
                    'properties': {
                        # 'errors': error_schema,
                        'fields': {
                            'type': 'object',
                            'properties': properties,
                            'required': required,
                            'additionalProperties': False,
                        },
                    },
                    'additionalProperties': False,
                    'required': [
                        'fields',
                        # 'errors',
                    ],
                },
            ],
        }

        return {
            '$ref': f'#/definitions/{definition_name}',
        }

    elif issubclass(Type, TypeHint):
        return None
        """
            return {
                'title': Type().name,
                'type': 'object',
                'additionalProperties': False,
           }
        """
    assert False


"""
def wrap_with_globals(props: Any, definitions: Dict) -> Any:
    message_schema = create_schema(Message, definitions)

    return {
        **props,
        'properties': {
            **props['properties'],
            'template_name': {'type': 'string'},
            'csrf_token': {'type': 'string'},
            'messages': {
                'type': 'array',
                'items': message_schema,
            },
        },
        'required': [
            *props['required'],
            'template_name',
            'csrf_token',
            'messages',
        ],
    }
"""


def generate_schema() -> str:
    definitions = {}

    schema = {
        'title': 'Schema',
        'type': 'object',
        'definitions': definitions,
        'properties': {
            # name: wrap_with_globals(create_schema(Props, definitions, ref=False), definitions)
            name: create_schema(Props, definitions, ref=False)
            for name, Props in type_registry.items()
        },
        'additionalProperties': False,
        'required': [name for name in type_registry.keys()],
    }

    return simplejson.dumps(schema, indent=4)


class WidgetType(TypeHint):
    name = 'WidgetType'


class FieldType(NamedTuple):
    name: str
    label: str
    help_text: str
    widget: WidgetType


FormErrors = Dict[str, Optional[List[str]]]


class FormType(NamedTuple):
    errors: FormErrors
    fields: Dict[str, FieldType]
    iterator: List[str]


class SSRFormRenderer:
    def render(self, template_name, context, request=None):
        return simplejson.dumps(context)


def serialize_form(form: Optional[forms.BaseForm]) -> Optional[FormType]:
    if form is None:
        return None

    form.renderer = SSRFormRenderer()
    fields = {
        field.name: FieldType(
            widget=simplejson.loads(str(field))['widget'],
            name=field.name,
            label=str(field.label), # This can be a lazy proxy, so we must call str on it.
            help_text=str(field.help_text), # This can be a lazy proxy, so we must call str on it.
        ) for field in form
    }

    return FormType(
        errors=form.errors,
        fields=fields,
        iterator=list(fields.keys()),
   )


def generate_settings() -> None:
    settings_to_serialize = {
        setting_name: setting_value
        for setting_name, setting_value in settings._wrapped.__dict__.items() if setting_name not in ['_explicit_settings']
    }
    return simplejson.dumps(settings_to_serialize, indent=4)



# Generated by Django 2.1b1 on 2018-07-10 17:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('testing', '0001_initial'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='Widget',
            new_name='Trinket',
        ),
    ]



from typing import Any

class Field:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

class CharField(Field): ...

class ChoiceField(Field): ...

class MultipleChoiceField(Field): ...

class ModelChoiceField(Field): ...

class ModelMultipleChoiceField(Field): ...

class FileField(Field): ...

class EmailField(Field): ...



from django.core.management.base import BaseCommand, CommandError

from typing import Any

from ... import generate_schema


class Command(BaseCommand):
    help = 'Generates a JSON schema for all registered types.'

    def handle(self, *args: Any, **options: Any) -> None:
        self.stdout.write(generate_schema())



class JSXTemplateMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        if getattr(request, '_is_jsx_response', False) is True:
            response['content-type'] = 'application/ssr+json'

        return response



from django.db import models

import decimal


class Category(models.Model):
    name: str = models.CharField(max_length=200)

    def __str__(self) -> str:
        return self.name


class Trinket(models.Model):
    name: str = models.CharField(max_length=200)
    category: Category = models.ForeignKey(Category, on_delete=models.CASCADE)
    price: decimal.Decimal = models.DecimalField(decimal_places=2, max_digits=10)
    type: str = models.CharField(max_length=100, choices=(
        ('cheap', 'Cheap'),
        ('expensive', 'Expensive'),
    ))

    def __str__(self) -> str:
        return self.name



from django.conf import settings
from django.middleware.csrf import get_token
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.template.backends.base import BaseEngine

import os
import simplejson
import markdown

from . import render_jsx


class JSX(BaseEngine):
    # Name of the subdirectory containing the templates for this engine
    # inside an installed application.
    # app_dirname = 'foobar'

    def __init__(self, params):
        # params = params.copy()
        options = params.pop('OPTIONS').copy()
        super().__init__(params)

        # self.engine = foobar.Engine(**options)

    def from_string(self, template_code):
        return 'ABC'
        """
        try:
          return Template(self.engine.from_string(template_code))
        except foobar.TemplateCompilationFailed as exc:
            raise TemplateSyntaxError(exc.args)
        """

    def get_template(self, template_name):
        if template_name.endswith('.html'):
            jsx_template_name = template_name.replace('.html', '.tsx')

            if os.path.isfile(os.path.join(settings.BASE_DIR, 'client/templates', jsx_template_name)):
                return Template(jsx_template_name)

        raise TemplateDoesNotExist([], backend=self)

        try:
            return Template(self.engine.get_template(template_name))
        except foobar.TemplateNotFound as exc:
            raise TemplateDoesNotExist(exc.args, backend=self)
        except foobar.TemplateCompilationFailed as exc:
            raise TemplateSyntaxError(exc.args)


class Template:
    def __init__(self, jsx_template_name):
        self.jsx_template_name = jsx_template_name

    def get_props(self, context=None, request=None):
        if self.jsx_template_name == 'registration/login.tsx':
            return {
                'form': context['form'],
            }
        elif self.jsx_template_name == '404.tsx':
            return {
                'request_path': context['request_path'],
            }
        elif self.jsx_template_name == 'flatpages/default.tsx':
            flatpage = context['flatpage']

            return {
                'flatpage': {
                    'title': flatpage.title,
                    'url': flatpage.url,
                    'content': markdown.markdown(flatpage.content),
                },
            }
        else:
            assert False


    def render(self, context=None, request=None):
        request._is_jsx_response = True

        props = self.get_props(context, request)
        template_name = self.jsx_template_name.replace('.tsx', '')

        return render_jsx(request, template_name, props)



#!/usr/bin/env python
import sys
import os

import reactivated

module_directory = os.path.dirname(reactivated.__file__)
conf_directory = os.path.join(module_directory, 'conf')

mypy = os.path.join(conf_directory, 'mypy.ini')
tsconfig = os.path.join(conf_directory, 'tsconfig.json')

def create_symlink(path: str, name: str) -> None:
    if os.path.isfile(name):
        os.remove(name)
    os.symlink(path, name)


create_symlink(mypy, 'mypy.ini')
create_symlink(mypy, 'mypy.ini')



# Generated by Django 2.0.7 on 2018-07-07 14:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='Widget',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('price', models.DecimalField(decimal_places=2, max_digits=10)),
                ('type', models.CharField(choices=[('cheap', 'Cheap'), ('expensive', 'Expensive')], max_length=100)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='testing.Category')),
            ],
        ),
    ]



from django.http import HttpRequest, HttpResponse
from django.middleware.csrf import get_token
from django.shortcuts import redirect

from typing import Any, NamedTuple, Generic, TypeVar, Union, Dict, Optional, List, Any, Tuple, Sequence, Mapping, overload

# from server.apps.testing import models

from . import ssr, generate_schema, TypeHint, FormType

import abc
import simplejson


class Trinket(NamedTuple):
    name: str
    url: str


class FormViewProps(NamedTuple):
    widget_list: List[Trinket]
    form: FormType



class SSRFormRenderer:
    def render(self, template_name, context, request=None):
        return simplejson.dumps(context)


def schema(request: HttpRequest) -> HttpResponse:
    return HttpResponse(generate_schema(), content_type='application/json')



type_registry: Dict[str, NamedTuple] = {}


class TrinketListProps(NamedTuple):
    trinket_list: List[Trinket]



from typing import Callable, Tuple, Type, cast
from mypy_extensions import TypedDict, Arg, KwArg


class EmptyParams(NamedTuple):
    pass


class TrinketListParams(NamedTuple):
    pass


@ssr(props=TrinketListProps, params=TrinketListParams)
def trinket_list(request: HttpRequest, params: TrinketListParams) -> TrinketListProps:
    trinket_list = [
        Trinket(
            name=trinket.name,
            url=f'/trinkets/{trinket.pk}/',
        ) for trinket in models.Trinket.objects.all()
    ]

    return TrinketListProps(
        trinket_list=trinket_list,
    )


class TrinketDetailParams(NamedTuple):
    pk: int


class TrinketDetailProps(NamedTuple):
    trinket: Trinket
    back_url: str


class NestedNewThing(NamedTuple):
    b: bool
    f: int


class NewThing(NamedTuple):
    a: str
    b: bool
    c: NestedNewThing


@ssr(props=TrinketDetailProps, params=TrinketDetailParams)
def trinket_detail(request: HttpRequest, params: TrinketDetailParams) -> TrinketDetailProps:
    trinket = models.Trinket.objects.get(pk=params.pk)

    return TrinketDetailProps(
        trinket=Trinket(
            name=trinket.name,
            url=f'/trinkets/{trinket.pk}/',
        ),
        back_url='/trinkets/',
    )


@ssr(props=NewThing, params=TrinketListParams)
def my_thang(request: HttpRequest, params: TrinketListParams) -> NewThing:
    return NewThing(
        a='a',
        b=True,
        c=NestedNewThing(
            b=True,
            f=5,
        ),
    )


@ssr(props=FormViewProps)
def form_view(request: HttpRequest) -> Union[FormViewProps, HttpResponse]:
    from django import forms

    class TrinketForm(forms.ModelForm):
        class Meta:
            model = models.Trinket
            fields = '__all__'

    if request.method == 'POST':
        form = TrinketForm(request.POST, renderer=SSRFormRenderer())

        if form.is_valid():
            form.save()
            return redirect(request.path)
    else:
        form = TrinketForm(renderer=SSRFormRenderer())

    serialized_form = FormType(
        errors=form.errors,
        fields=[
            FieldType(
                widget=simplejson.loads(str(field))['widget'],
                name=field.name,
                label=field.label,
            ) for field in form
        ],
   )

    return FormViewProps(
        form=serialized_form,
        widget_list=[
            Trinket(
                name=trinket.name,
                url=f'/trinkets/{trinket.pk}/',
            ) for trinket in models.Trinket.objects.all()
        ],
    )



from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from typing import Any

from ... import generate_settings


class Command(BaseCommand):
    help = 'Generates a JSON representation of all settings.'

    def handle(self, *args: Any, **options: Any) -> None:
        self.stdout.write(generate_settings())



import setuptools
import os

def find_stubs(package):
    stubs = []
    for root, dirs, files in os.walk(package):
        for file in files:
            path = os.path.join(root, file).replace(package + os.sep, '', 1)
            stubs.append(path)
    return {package: stubs}

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="reactivated",
    version="0.0.6",
    author="Silvio Gutierrez",
    author_email="silviogutierrez@gmail.com",
    description="A statically typed framework to create Django sites with a React frontend.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/silviogutierrez/django-react",
    packages=setuptools.find_packages(),
    package_data={
        'reactivated': [
            'conf/mypy.ini',
            'py.typed',
        ],
        **find_stubs('django-stubs'),
        **find_stubs('custom_user-stubs'),
    },
    scripts=[
        'scripts/to-be-renamed.py',
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



from typing import Any, Dict, List, Optional, Iterator

class BaseForm:
    cleaned_data: Dict[str, Any]

    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

    def is_valid(self) -> bool: ...

    errors: Dict[str, Optional[List[str]]]

    def __iter__(self) -> Iterator[Any]: ...

class Form(BaseForm): ...

class ModelForm(Form):
    def save(self, commit=False) -> Any: ...
